-- Lua optimizations
local ceil = ceil
local date = date
local format = format
local gsub = gsub
local strfind = strfind
local strlen = strlen
local strsplit = strsplit
local strsub = strsub
local strupper = strupper
local time = time
local tonumber = tonumber
local GetRealmName = GetRealmName
local GetNumSavedInstances = GetNumSavedInstances
local GetSavedInstanceInfo = GetSavedInstanceInfo
local SecondsToTime = SecondsToTime
local RequestRaidInfo = RequestRaidInfo
local UnitClass = UnitClass
local UnitName = UnitName

-- Use Blizzard's strings to save localization efforts
local NO_RAID_INSTANCES_SAVED = NO_RAID_INSTANCES_SAVED -- "You are not saved to any instances"

local ICON_HEROIC = ICON_LIST[8].."16:16:0:0|t" -- Skull raid target icon
local ICON_NORMAL = ICON_LIST[1].."16:16:0:0|t" -- Star raid target icon

local SavedInstances = LibStub("AceAddon-3.0"):NewAddon("SavedInstances", "AceEvent-3.0")
_G.SavedInstances = SavedInstances
local Config = LibStub("AceConfig-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SavedInstances")

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)
local launcher

local QTip = LibStub("LibQTip-1.0")
local tooltip

local selectedInstance -- for re-ordering the instance list

local db
local defaultDB = {
	DBVersion = 2,
	Toons = { }, 	-- table key: "Toon - Realm"; value:
				-- Class: string
				-- Hide: boolean
				-- NumSaved: integer
	Tooltip = {
		FreshToons = false,
		AllRaids = true,
		Details = false,
	},
	Instances = { }, 	-- table key: "Instance name"; value:
					-- Show: boolean
					-- Order: integer
					-- table key: "Toon - Realm"; value:
						-- table key: "Difficulty"; value:
							-- ID: integer
							-- Expires: integer
	FuBar = {
		AttachMinimap = false,
		HideMinimapButton = true,
	},
	MinimapIcon = { },
	profile = { },	-- LibFuBarPlugin-3.0 expects this /sigh
}

-- general helper functions below

local function ClassColorise(class, targetstring)
	local class = strupper(gsub(class, " ", ""))
	local r = format("%x", math.floor(RAID_CLASS_COLORS[class].r * 255))
	if strlen(r) == 1 then
		r = "0"..r
	end
	local g = format("%x", math.floor(RAID_CLASS_COLORS[class].g * 255))
	if strlen(g) == 1 then
		g = "0"..g
	end
	local b = format("%x", math.floor(RAID_CLASS_COLORS[class].b * 255))
	if strlen(b) == 1 then
		b = "0"..b
	end
	return "|cff"..r..g..b..targetstring..FONT_COLOR_CODE_CLOSE
end

local function TableLen(table)
	local i = 0
	local k, v
	for k, v in pairs(table) do
		i = i + 1
	end
	return i
end

local function OrderedInstances()
	local list = { }
	local instance, a
	for instance, a in pairs(SavedInstances.db.Instances) do
		list[a.Order] = instance
	end
	return list
end

local function SetInstanceOrder(from, to)
	local k, v
	for k, v in pairs(SavedInstances.db.Instances) do
		if v.Order == from then
			v.Order = to
		elseif to < from and v.Order >= to and v.Order < from then
			v.Order = v.Order + 1
		elseif to > from and v.Order <= to and v.Order > from then
			v.Order = v.Order - 1
		end
	end
end

-- addon code below

function SavedInstances:OnInitialize()
	local toon = UnitName("player").." - "..GetRealmName()
	SavedInstancesDB = SavedInstancesDB or defaultDB
	if not SavedInstancesDB.DBVersion or SavedInstancesDB.DBVersion ~= 2 then
		SavedInstancesDB = defaultDB
	end
	self.db = db or SavedInstancesDB
	self.db.Toons[toon] = self.db.Toons[toon] or { }
	self.db.Toons[toon].Class = self.db.Toons[toon].Class or select(2, UnitClass("player"))
	self.db.Toons[toon].Hide = self.db.Toons[toon].Hide or false
	self.db.Toons[toon].NumSaved = self.db.Toons[toon].NumSaved or 0
	SavedInstances:SetupOptions()
	if LDB then
		self.launcher = LDB:NewDataObject("SavedInstances", {
			label = "SavedInstances",
			type = "launcher",
			icon = "Interface\\Addons\\SavedInstances\\icon.tga",
			OnClick = function(clickedframe, button)
				self:ShowConfig()
			end,
			OnEnter = function(tt)
				SavedInstances:ShowTooltip(tt)
			end,
			OnLeave = SavedInstances.HideTooltip
		})
	end
	if LDBIcon then
		LDBIcon:Register("SavedInstances", self.launcher, self.db.MinimapIcon)
	end
	if LibStub:GetLibrary("LibFuBarPlugin-3.0", true) and not IsAddOnLoaded("FuBar2Broker") then
		local LFBP = LibStub:GetLibrary("LibFuBarPlugin-3.0")
		LibStub("AceAddon-3.0"):EmbedLibrary(self, "LibFuBarPlugin-3.0")
		self:SetFuBarOption("tooltipType", "Custom")
		self:SetFuBarOption("hasNoColor", true)
		self:SetFuBarOption("cannotDetachTooltip", true)
		self:SetFuBarOption("hideWithoutStandby", true)
		self:SetFuBarOption("iconPath", "Interface\\Addons\\SavedInstances\\icon.tga")
		self:SetFuBarOption("hasIcon", true)
		self:SetFuBarOption("defaultPosition", "RIGHT")
		self:SetFuBarOption("tooltipHiddenWhenEmpty", false)
		self:SetFuBarOption("configType", "None")
		LFBP:OnEmbedInitialize(self)
		self:GetFrame():SetScript("OnEnter", function()
			SavedInstances:ShowTooltip(SavedInstances:GetFrame())
		end)
		self:GetFrame():SetScript("OnLeave", SavedInstances.HideTooltip)
		function SavedInstances:OnFuBarClick(button)
				self:ShowConfig()
		end
		self:UpdateFuBarSettings()
		self.optionsFrames.FuBar = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", "FuBar Options", "SavedInstances", "FuBar")
	end
	self:RequestRefresh()
end

function SavedInstances:OnEnable()
	self:RegisterEvent("UPDATE_INSTANCE_INFO", "Refresh")
	self:RegisterEvent("RAID_INSTANCE_WELCOME", "RequestRefresh")
	self:RegisterEvent("CHAT_MSG_SYSTEM", "CheckSystemMessage")
end

function SavedInstances:OnDisable()
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("RAID_INSTANCE_WELCOME")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")
end

function SavedInstances.HideTooltip()
	QTip:Release(SavedInstances.tooltip)
	SavedInstances.tooltip = nil
end

function SavedInstances:ShowTooltip(parentframe)
	SavedInstances.tooltip = QTip:Acquire("SavedInstancesTooltip", 1, "LEFT")
	local tooltip = SavedInstances.tooltip
	local thisToon = UnitName("player").." - "..GetRealmName()
	local hFont = tooltip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	tooltip:SetHeaderFont(hFont)
	self:CullStaleInstances()
	local columns = { }
	if not self.db.Toons[thisToon].Hide
			and ((self.db.Toons[thisToon].NumSaved == 0 and self.db.Tooltip.FreshToons)
			or self.db.Toons[thisToon].NumSaved > 0) then
		columns[thisToon] = tooltip:AddColumn("CENTER")
	end
	local headLine, headCol = tooltip:AddHeader(NORMAL_FONT_COLOR_CODE.."SavedInstances"..FONT_COLOR_CODE_CLOSE)
	local i, instance
	for i, instance in ipairs(OrderedInstances()) do
		local currInstance = self.db.Instances[instance]
		if TableLen(currInstance) > 2 then -- instance has toon info
			local currLine, _ = tooltip:AddLine(GRAY_FONT_COLOR_CODE..instance..FONT_COLOR_CODE_CLOSE)
			local toon, v
			for toon, v in pairs(self.db.Toons) do
				if currInstance[toon] and not self.db.Toons[toon].Hide then
					tooltip:SetCell(currLine, 1, NORMAL_FONT_COLOR_CODE..instance..FONT_COLOR_CODE_CLOSE)
					if not columns[toon] then
						columns[toon] = tooltip:AddColumn("CENTER")
					end	
					local diffString = ""
					if currInstance[toon]["Normal"] then
						diffString = ICON_NORMAL
						if IsShiftKeyDown() then
							diffString = diffString..tostring(currInstance[toon]["Normal"].ID)
						elseif IsControlKeyDown() and not (currInstance[toon]["Heroic"] and currInstance[toon]["Normal"].Expires == currInstance[toon]["Heroic"].Expires) then
							diffString = diffString..SecondsToTime(currInstance[toon]["Normal"].Expires - time())
						end
					end
					if currInstance[toon]["Heroic"] then
						diffString = diffString..ICON_HEROIC
						if IsShiftKeyDown() then
							diffString = diffString..tostring(currInstance[toon]["Heroic"].ID)
						elseif IsControlKeyDown() then
							diffString = diffString..SecondsToTime(currInstance[toon]["Heroic"].Expires - time())
						end
					end
					tooltip:SetCell(currLine, columns[toon], diffString)
				end
			end
		elseif self.db.Tooltip.AllRaids and currInstance.Show then
			-- instance has no toon info, but must be shown anyway
			tooltip:AddLine(GRAY_FONT_COLOR_CODE..instance..FONT_COLOR_CODE_CLOSE)
		end
	end
	if tooltip:GetLineCount() == 1 then
		tooltip:AddLine(GRAY_FONT_COLOR_CODE..NO_RAID_INSTANCES_SAVED..FONT_COLOR_CODE_CLOSE)
	end
	if self.db.Tooltip.FreshToons then
		local toon, v
		for toon, v in pairs(self.db.Toons) do
			if not self.db.Toons[toon].Hide and not columns[toon] then
				columns[toon] = tooltip:AddColumn("CENTER")
			end
		end
	end
	local toon, col
	for toon, col in pairs(columns) do
		tooltip:SetCell(headLine, col, ClassColorise(self.db.Toons[toon].Class, strsplit(' ', toon)), tooltip:GetHeaderFont(), "CENTER")
	end 
	tooltip:AddLine(" ")
	local hintLine, hintCol
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, L["|cffffff00Click|r to configure"], "LEFT", tooltip:GetColumnCount())
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, L["|cffffff00Shift|r for IDs, |cffffff00Control|r for time left"], "LEFT", tooltip:GetColumnCount())
	tooltip:SmartAnchorTo(parentframe)
	tooltip:Show()
end

function SavedInstances:CheckSystemMessage()
	if tostring(arg1) == INSTANCE_SAVED then
		self:RequestRefresh()
	end
end

function SavedInstances:RequestRefresh()
	RequestRaidInfo()
end

function SavedInstances:Refresh()
	local toon = UnitName("player").." - "..GetRealmName()
	self:CullStaleInstances()
	self.db.Toons[toon].NumSaved = GetNumSavedInstances()
	if self.db.Toons[toon].NumSaved > 0 then
		local i, iname, iid, iexpires, idiff, temp
		for i = 1, self.db.Toons[toon].NumSaved do
			iname, iid, iexpires, idiff = GetSavedInstanceInfo(i)
			if idiff > 1 then
				idiff = "Heroic"
			else
				idiff = "Normal"
			end
			self.db.Instances[iname] = self.db.Instances[iname] or { }
			self.db.Instances[iname].Order = self.db.Instances[iname].Order or TableLen(self.db.Instances)
			self.db.Instances[iname].Show = self.db.Instances[iname].Show or true
			self.db.Instances[iname][toon] = self.db.Instances[iname][toon] or { }
			self.db.Instances[iname][toon][idiff] = {
				ID = iid,
				Expires = iexpires + time()
			}
		end
	end
end

function SavedInstances:CullStaleInstances()
	local instance, a, toon, b, difficulty, c
	for instance, a in pairs(SavedInstances.db.Instances) do
		for toon, b in pairs(a) do
			if type(b) == "table" then
				for difficulty, c in pairs(b) do
					if c.Expires < time() then
						SavedInstances.db.Instances[instance][toon][difficulty] = nil
					end
				end
				if TableLen(b) == 0 then
					SavedInstances.db.Instances[instance][toon] = nil
				end
			end
		end
	end
end

-- options stuff below

function SavedInstances:UpdateFuBarSettings()
	if LibStub:GetLibrary("LibFuBarPlugin-3.0", true) then
		if SavedInstances.db.FuBar.HideMinimapButton then
			self:Hide()
		else
			self:Show()
			if self:IsFuBarMinimapAttached() ~= SavedInstances.db.FuBar.AttachMinimap then
				self:ToggleFuBarMinimapAttached()
			end
		end
	end
end

local function GetFuBarMinimapAttachedStatus(info)
	return SavedInstances:IsFuBarMinimapAttached() or SavedInstances.db.FuBar.HideMinimapButton
end

local options = {
	type = "group",
	name = "SavedInstances",
	handler = SavedInstances,
	args = {
		General = {
			order = 1,
			type = "group",
			name = L["General settings"],
			desc = L["General settings"],
			args = {
				intro = {
					order = 1,
					type = "description",
					name = L["Track the instance IDs saved against your characters"],
				},
				MinimapGroup = {
					order = 2,
					type = "group",
					guiInline = true,
					name = L["Minimap button"],
					desc = L["Minimap button"],
					hidden = function() return not LDBIcon end,
					args = {
						note = {
							order = 1,
							type = "description",
							name = L["Note: this is the recommended control for the minimap icon"],
						},
						MinimapIcon = {
							type = "toggle",
							name = L["Show minimap button"],
							desc = L["Show the SavedInstances minimap button"],
							order = 2,
							get = function(info) return not SavedInstances.db.MinimapIcon.hide end,
							set = function(info, value)
								SavedInstances.db.MinimapIcon.hide = not value
								if value then LDBIcon:Show("SavedInstances") else LDBIcon:Hide("SavedInstances") end
							end,
						},
					},
				},
				ToonsGroup = {
					order = 3,
					type = "group",
					guiInline = true,
					name = L["Character options"],
					desc = L["Character options"],
					args = {
						FreshToons = {
							type = "toggle",
							name = L["Show non-saved characters"],
							desc = L["Show characters that have not been saved to any instances"],
							order = 1,
							get = function(value)
									return SavedInstances.db.Tooltip.FreshToons
							end,
							set = function(info, value)
									SavedInstances.db.Tooltip.FreshToons = value
							end,
						},
						intro = {
							order = 2,
							type = "description",
							name = L["Select which detected characters to never show in the tooltip"],
						},
						List = {
							order = 3,
							type = "multiselect",
							name = L["Ignore these characters"],
		--					desc = "Never show these characters in the tooltip",
							get = function(info, toon)
								return SavedInstances.db.Toons[toon].Hide
							end,
							set = function(info, toon, value)
								SavedInstances.db.Toons[toon].Hide = value
							end,
							values = function()
								local list = { }
								local k, v
								for k, v in pairs(SavedInstances.db.Toons) do
									list[k] = k
								end
								return list
							end,
						},
					},
				},
			},
		},
		Instances = {
			order = 4,
			type = "group",
			name = L["Instance options"],
			desc = L["Instance options"],
			args = {
				OrderGroup = {
					order = 1,
					type = "group",
					guiInline = true,
					name = L["Instance order"],
					desc = L["Instance order"],
					args = {
						Selected = {
							order = 1,
							type = "select",
							name = L["Instance to be moved"],
							get = function(info, instance)
								return SavedInstances.selectedInstance
							end,
							set = function(info, instance)
								SavedInstances.selectedInstance = instance
							end,
							values = OrderedInstances,
						},
						Position = {
							order = 2,
							type = "select",
							name = L["Position in list"],
							get = function(info, instance)
								return SavedInstances.selectedInstance
							end,
							set = function(info, instance)
								SetInstanceOrder(SavedInstances.selectedInstance, instance)
								SavedInstances.selectedInstance = nil
							end,
							values = OrderedInstances,
						},
					},
				},
				VisibilityGroup = {
					order = 2,
					type = "group",
					guiInline = true,
					name = L["Instance visibility"],
					desc = L["Instance visibility"],
					args = {
						AllRaids = {
							type = "toggle",
							name = L["Show all instances"],
							desc = L["Always show all instances, even if no characters are saved to them"],
							order = 1,
							get = function(value)
									return SavedInstances.db.Tooltip.AllRaids
							end,
							set = function(info, value)
									SavedInstances.db.Tooltip.AllRaids = value
							end,
						},
						List = {
							order = 2,
							type = "multiselect",
							name = L["Show these instances when all instances are shown"],
							get = function(info, instance)
								local trimmedInstance = strsub(instance, 5)
								return SavedInstances.db.Instances[trimmedInstance].Show
							end,
							set = function(info, instance, value)
								local trimmedInstance = strsub(instance, 5)
								SavedInstances.db.Instances[trimmedInstance].Show = value
							end,
							values = function()
								local list = { }
								local i, v
								for i, v in ipairs(OrderedInstances()) do
									local prefix = format("%04d", i)
									list[prefix..v] = v
								end
								return list
							end,
						},
					},
				},
			},
		},
		FuBar = {
			order = -1,
			type = "group",
			name = "FuBar Options",
			desc = "FuBar Options",
			hidden = function() return SavedInstances.IsFuBarMinimapAttached == nil end,
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Configure FuBar and related options",
				},
				hideIcon = {
					type = "toggle",
					order = 2,
					name = "Hide minimap/FuBar icon",
					desc = "Hide minimap/FuBar icon",
					get = function(info) return SavedInstances.db.FuBar.HideMinimapButton end,
					set = function(info, v)
						SavedInstances.db.FuBar.HideMinimapButton = v
						SavedInstances:UpdateFuBarSettings()
					end,
				},
				showIcon = {
					type = "toggle",
					order = 3,
					name = "Show icon",
					desc = "Show icon",
					get = function(info) return SavedInstances:IsFuBarIconShown() end,
					set = function(info, v) SavedInstances:ToggleFuBarIconShown() end,
					disabled = GetFuBarMinimapAttachedStatus,
				},
				showText = {
					type = "toggle",
					order = 4,
					name = "Show text",
					desc = "Show text",
					get = function(info) return SavedInstances:IsFuBarTextShown() end,
					set = function(info, v) SavedInstances:ToggleFuBarTextShown() end,
					disabled = GetFuBarMinimapAttachedStatus,
				},
				position = {
					type = "select",
					order = 5,
					name = "Position",
					desc = "Position",
					values = {LEFT = "Left", CENTER = "Center", RIGHT = "Right"},
					get = function() return SavedInstances:GetPanel() and SavedInstances:GetPanel():GetPluginSide(SavedInstances) end,
					set = function(info, val)
						if SavedInstances:GetPanel() and SavedInstances:GetPanel().SetPluginSide then
							SavedInstances:GetPanel():SetPluginSide(SavedInstances, val)
						end
					end,
					disabled = GetFuBarMinimapAttachedStatus,
				},
				MinimapGroup = {
					order = 6,
					type = "group",
					guiInline = true,
					name = "Minimap button",
					desc = "Minimap button",
					args = {
						note = {
							order = 1,
							type = "description",
							name = "Note: It is recommended you leave the minimap control here disabled.",
						},
						attachMinimap = {
							type = "toggle",
							order = 2,
							name = "Attach to minimap",
							desc = "Attach to minimap",
							get = function(info) return SavedInstances:IsFuBarMinimapAttached() end,
							set = function(info, v)
								SavedInstances:ToggleFuBarMinimapAttached()
								SavedInstances.db.FuBar.AttachMinimap = SavedInstances:IsFuBarMinimapAttached()
							end,
							disabled = function() return SavedInstances.db.FuBar.HideMinimapButton end,
						},
					},
				},
			},
		},
	},
}
SavedInstances.Options = options

function SavedInstances:SetupOptions()
	self.optionsFrames = {}
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("SavedInstances", options)
	local ACD3 = LibStub("AceConfigDialog-3.0")
	self.optionsFrames.SavedInstances = ACD3:AddToBlizOptions("SavedInstances", nil, nil, "General")
	self.optionsFrames.Instances = ACD3:AddToBlizOptions("SavedInstances", L["Instance options"], "SavedInstances", "Instances")
end

function SavedInstances:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.Instances)
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.SavedInstances)
end
